#include <string.h>
#include <DFRobot_ENS160.h>
#include "DFRobot_BME280.h"
//inputs ----------------------------
#define LED 13
#define COMSPIN 27

//*************WHO?********************
uint8_t BOXID = 1;
//*************************************

#define transistorPinSensors 23
#define SEA_LEVEL_PRESSURE 1038.0f
#define LDR_PIN 33

bool sensorSetupFinished = false;

//sensors adress------------------------
DFRobot_ENS160_I2C ENS160(&Wire, /*i2cAddr*/ 0x53);
typedef DFRobot_BME280_IIC BME;  // ******** use abbreviations instead of full names ********
BME bme(&Wire, 0x76);            // select TwoWire peripheral and set sensor address

//ESP OS
//wifi--------------------------------------------
#include <WiFi.h>
#include <HTTPClient.h>
#include <WebServer.h>
#include <Preferences.h>


//esp name and credentials ------------------------
const char* name = "B.I.R.D.";  //your name
char systemName[20];            //device name

//variables for writing our credentials to memmory
Preferences preferences;  //setting up memmory management
String ssid, password;    //setting up Strings

//webserver-------------------------------------------------
WebServer server(80);              //specifying the webserver port
bool internetSetupActive = false;  //is the webserver already in setupmode
String responseHeader;             //specifying a place where the users response is stored


void setup() {
  pinMode(LDR_PIN, INPUT);
  Serial.begin(115200);

  //Enviromental sensor
  pinMode(transistorPinSensors, OUTPUT);
  
  // the AI saw a thing input
  pinMode(COMSPIN, INPUT);

  delay(500);
  wifiManagerSetup();
}

void loop() {
  if (internetSetupActive) {
    server.handleClient();
  }else{
    if(digitalRead(COMSPIN)==HIGH){
      readSensorData(true);
    }
  }
}

// show last sensor operate status
void printLastOperateStatus(BME::eStatus_t eStatus) {
  switch (eStatus) {
    case BME::eStatusOK: Serial.println("everything ok"); break;
    case BME::eStatusErr: Serial.println("unknow error"); break;
    case BME::eStatusErrDeviceNotDetected: Serial.println("device not detected"); break;
    case BME::eStatusErrParameter: Serial.println("parameter error"); break;
    default: Serial.println("unknow status"); break;
  }
}

void setupEnviromentalSensor() {

  bme.reset();
  Serial.println("bme read data test");
  while (bme.begin() != BME::eStatusOK) {
    Serial.println("bme begin faild");
    printLastOperateStatus(bme.lastOperateStatus);
    delay(2000);
  }
  Serial.println("bme begin success");
  delay(100);

  // Init the sensor
  while (NO_ERR != ENS160.begin()) {
    Serial.println("Communication with device failed, please check connection");
    delay(3000);
  }

  ENS160.setPWRMode(ENS160_STANDARD_MODE);

  ENS160.setTempAndHum(/*temperature=*/bme.getTemperature(), /*humidity=*/bme.getHumidity());
}

void readSensorData(bool wasThereABird) {
  delay(1000);
  digitalWrite(transistorPinSensors, HIGH);

  setupEnviromentalSensor();
  delay(120000);
  //LDR readigs
  int ldrValue = analogRead(LDR_PIN);
  int mappedLDR = map(ldrValue, 0, 3000, 100, 1);
  if (mappedLDR < 1) {
    mappedLDR = 1;
  }
  //Enviromental sensor readings
  float temp = bme.getTemperature();
  uint32_t press = bme.getPressure();
  float alti = bme.calAltitude(SEA_LEVEL_PRESSURE, press);
  float humi = bme.getHumidity();
  /**
   * Get the air quality index
   * Return value: 1-Excellent, 2-Good, 3-Moderate, 4-Poor, 5-Unhealthy
   */
  uint8_t AQI = ENS160.getAQI();
  /**
   * Get TVOC concentration
   * Return value range: 0–65000, unit: ppb
   */
  uint16_t TVOC = ENS160.getTVOC();
  /**
   * Get CO2 equivalent concentration calculated according to the detected data of VOCs and hydrogen (eCO2 – Equivalent CO2)
   * Return value range: 400–65000, unit: ppm
   * Five levels: Excellent(400 - 600), Good(600 - 800), Moderate(800 - 1000), 
   *               Poor(1000 - 1500), Unhealthy(> 1500)
   */
  uint16_t ECO2 = ENS160.getECO2();

  Serial.println();
  Serial.println("======== start print ========");
  Serial.print("temperature (unit Celsius): ");
  Serial.println(temp);
  Serial.print("pressure (unit pa):         ");
  Serial.println(press);
  Serial.print("altitude (unit meter):      ");
  Serial.println(alti);
  Serial.print("humidity (unit percent):    ");
  Serial.println(humi);
  Serial.print("light (unit percent):       ");
  Serial.println(mappedLDR);
  Serial.println("=============================");
  Serial.print("Air quality index :         ");
  Serial.println(AQI);
  Serial.print("Concentration VOC :         ");
  Serial.print(TVOC);
  Serial.println(" ppb");
  Serial.print("CO2 concentration :         ");
  Serial.print(ECO2);
  Serial.println(" ppm");
  Serial.println("========  end print  ========");
  Serial.println(" ");


  digitalWrite(transistorPinSensors, LOW);
  delay(500);

  post("https://somewebsite.lmao.com", uint8_t(mappedLDR), int8_t(AQI), double(temp), long(humi), press, TVOC, ECO2, wasThereABird, BOXID);

  delay(1000);
}


void post(String url, uint8_t LDR, int8_t AirQuality, double Temperature, long Humidity, uint32_t Pressure, uint16_t VOC, uint16_t CO2, bool isDetected, int8_t boxID) {

  if (WiFi.status() == WL_CONNECTED) {  //Check WiFi connection status
    HTTPClient http;

    Serial.println("Commence communication");
    String postData = "addVariables=1";
    postData += "&LDR=" + String(LDR);
    postData += "&AirQuality=" + String(AirQuality);
    postData += "&Temperature=" + String(Temperature);
    postData += "&Humidity=" + String(Humidity);
    postData += "&Pressure=" + String(Pressure);
    postData += "&VOC=" + String(VOC);
    postData += "&CO2=" + String(CO2);
    postData += "&Detected=" + String(isDetected);
    postData += "&boxID=" + String(boxID);

    http.begin(url);
    http.addHeader("Content-Type", "application/x-www-form-urlencoded");
    int httpResponseCode = http.POST(postData);
    Serial.println(postData);
    String response = http.getString();
    Serial.println(httpResponseCode);
    Serial.println(response);

    http.end();
    delay(5000);
  }
}


void wifiManagerSetup() {
  //setting up our system
  strcpy(systemName, "ESP_OS_");
  strcat(systemName, name);

  delay(100);

  //bootup saying hello
  Serial.println();
  Serial.println("ESP_OS");
  Serial.print("Welcome, ");
  Serial.println(name);

  delay(10);

  //starting our trials to connect to a wifi network
  Serial.println("Attempting to connect to wifi.");
  preferences.begin("ESP_OS_", true);  //setting up writing to memmory

  if (preferences.getString("ssid", "NULL") != "NULL" && preferences.getString("password", "NULL") != "NULL") {
    //wifi pass and login has been found in memmory

    //retrieve them
    password = preferences.getString("password", "NULL");
    ssid = preferences.getString("ssid", "NULL");
    preferences.end();

    //start connection with wifi network
    Serial.println("Starting connection with " + ssid);

    wifiStartup();

  } else {
    //there hasn't been any wifi credentials, so booting up in AP mode...
    Serial.println("No valid string found, starting wifi setup.");
    preferences.end();
    wifiSetup();
  }
}

void wifiStartup() {
  WiFi.mode(WIFI_STA);
  Serial.print("Name: ");
  Serial.println(ssid.c_str());
  Serial.print("Password: ");
  Serial.println(password.c_str());
  WiFi.begin(ssid.c_str(), password.c_str());
  Serial.print("Connecting to WiFi ..");

  //if the wifi doesnt connect for to long, we go ahead and start up our
  //wifisetup wizard again
  int wifi_idle_counter = 0;
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print('.');
    if (wifi_idle_counter >= 16) {
      Serial.println();
      Serial.println("returning to setup");
      wifiSetup();
      return;
    }
    delay(1000);
    wifi_idle_counter++;
  }
  Serial.println();

  //if it does connect, great!
  Serial.println("succesfully connected!");
  Serial.print("Ip Adress: ");
  Serial.println(WiFi.localIP());
}

void wifiSetup() {
  //stop the current wifi nonsense
  WiFi.softAPdisconnect();
  WiFi.disconnect();
  server.stop();

  //spin up a temporary webserver
  WiFi.mode(WIFI_AP);

  Serial.print("Join the network: ");
  Serial.println(systemName);
  WiFi.softAP(systemName);
  WiFi.softAPConfig(IPAddress(192, 168, 4, 1), IPAddress(192, 168, 4, 1), IPAddress(255, 255, 255, 0));

  IPAddress IP = WiFi.softAPIP();
  Serial.print("AP IP address: ");
  Serial.println(IP);
  internetSetupActive = true;
  server.begin();
  Serial.println("Server has started, /\\ visit tis ip  ");

  //setting up the methods for the server to handle,
  //if someone visits ipadres/submit handlesubmit is called
  server.on("/", handle_OnConnect);
  server.on("/submit", handle_submit);
  server.onNotFound(handle_NotFound);
}

//handles when a user request a connection
void handle_OnConnect() {
  Serial.println("Webclient has returned a connect request.");
  server.send(200, "text/html", SendHTML(0));
}

//handles when a requested url is not found
void handle_NotFound() {
  server.send(404, "text/plain", "The webpage was not found");
}

//this method gets called once someone presses on the
//submit button on our webpage, where they added in new credentials.
void handle_submit() {
  String wifiName_ = server.arg("WifiName");  //check what is returned as our wifiname
  String password_ = server.arg("Password");  //check what is returned as our password

  if (wifiName_ != "") {  //if the wifiname is not empty we run the code, because an empty password can be true
    Serial.print("Specified wifi name: ");
    Serial.println(wifiName_);
    Serial.print("Specified wifi password: ");
    Serial.println(password_);

    //display the visitor of our webpage that their credentials have been excepted
    server.send(200, "text/html", SendHTML(1));

    //saving these credentials to flash
    preferences.begin("ESP_OS_", false);
    preferences.putString("password", password_);
    preferences.putString("ssid", wifiName_);
    password = preferences.getString("password", "NULL");
    ssid = preferences.getString("ssid", "NULL");
    preferences.end();
    Serial.println("All safely stored!");

    //stop the current wifi and webserver nonsense
    WiFi.softAPdisconnect();
    WiFi.disconnect();
    server.stop();
    Serial.println("Starting connection with: " + password);

    internetSetupActive = false;
    delay(100);

    //and startup the wifi
    wifiStartup();

  } else {  //if the wifiname is not empty send the normal webpage again!
    server.send(200, "text/html", SendHTML(0));
  }
}

//html things!
String SendHTML(uint8_t caseNum) {
  //document header
  String ptr = "<!DOCTYPE html> <html>\n";
  ptr += "<head><meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0, user-scalable=no\">\n";
  ptr += "<title>ESP_OS</title>\n";

  //styling
  ptr += "<style>";
  ptr += "h1,h3{color:#eee}.button,p{font-size:16px}html{font-family:Helvetica;display:inline-block;margin:0 auto;text-align:center}body{margin-top:50px;background-color:#282828}h1{margin:50px auto 30px}h3{margin-bottom:50px}p{color:#e9e9e9;margin-bottom:10px}.button{display:block;width:160px;background-color:#3498db;border:none;color:#fff;padding:13px 30px;text-decoration:none;margin:20px auto 35px;cursor:pointer;border-radius:2px}.button-on,.button-on:active,input[type=text]{background-color:#5f4bb6}.button-off{background-color:#3a325a}.button-off:active{background-color:#3a2c50}input[type=text]{width:12rem;padding:12px 20px;margin:8px 10px;font-size:14px;box-sizing:border-box;color:#e9e9e9;border:none;border-radius:2px}";
  ptr += "</style>\n";
  ptr += "</head>\n";

  //html dependent on whats happening on our server
  switch (caseNum) {
    case 0:
      ptr += "<body>\n";
      ptr += "<h1>ESP_OS WifiWizard</h1>\n";
      ptr += "<h3>Welcome ";
      ptr += name;
      ptr += "</h3>\n";

      //input fields
      ptr += "<form action = \"/submit\" method = \"get\">";
      ptr += "<input type = \"text\" name = \"WifiName\" id = \"WifiName\"></input>";
      ptr += "<input type = \"text\" name = \"Password\" id = \"Password\"></input>";
      ptr += "<input type = \"submit\" class=\"button button-on\" value = \"Submit\"></input>";
      ptr += "</form>";

      // ptr += "<a class=\"button button-on\" href=\"/submit\">Submit</a>\n";

      break;

    case 1:
      ptr += "<body>\n";
      ptr += "<h1>ESP_OS WifiWizard</h1>\n";
      ptr += "<h3>Welcome ";
      ptr += name;
      ptr += "</h3>\n";
      ptr += "<p>Input has been accepted and the ESP will try to connect shortly";
      ptr += "<p>\n";
      break;

    case 2:
      break;

    default:
      break;
  }

  //closing html tags...
  ptr += "</body>\n";
  ptr += "</html>\n";
  return ptr;
}
